# -*- coding: utf-8 -*-
# vim: sw=4:ts=4:expandtab

"""
examples
~~~~~~~~

README examples

Examples:
    Setup

        >>> import sys
        >>> from json import loads
        >>> from io import StringIO
        >>>
        >>> sys.stderr = sys.stdout


    Hello World

        >>> from pygogo import logger
        >>>
        >>> logger.debug('hello world')
        hello world
        >>> logger.error('hello error')
        hello error
        hello error


    Log based debugging

        >>> import pygogo as gogo
        >>>
        >>> def main(name, verbose=False):
        ...     logger = gogo.Gogo(name, verbose=verbose).logger
        ...     logger.debug('I will log to `stdout` if `verbose` is True')
        ...     logger.info('I will log to `stdout` always')
        ...     logger.warning('I will log to both `stdout` and `stderr`')
        >>>
        >>> main('quite')
        I will log to `stdout` always
        I will log to both `stdout` and `stderr`
        I will log to both `stdout` and `stderr`
        >>> main('verbose', True)
        I will log to `stdout` if `verbose` is True
        I will log to `stdout` always
        I will log to both `stdout` and `stderr`
        I will log to both `stdout` and `stderr`


    Disabled dual logging

        >>> import pygogo as gogo
        >>>
        >>> logger = gogo.Gogo('monolog', monolog=True).logger
        >>> logger.debug('debug message')
        debug message
        >>> logger.info('info message')
        info message
        >>> logger.warning('warning message')
        warning message
        >>> logger.error('error message')
        error message
        >>> logger.critical('critical message')
        critical message


    Using LoggerAdapters to impart contextual information

        >>> import pygogo as gogo
        >>>
        >>> s = StringIO()
        >>> going = gogo.Gogo('context', low_hdlr=gogo.handlers.fileobj_hdlr(s))
        >>> logger = going.get_structured_logger(connid='1234')
        >>> logger.info('log message')
        >>> loads(s.getvalue()) == {'message': 'log message', 'connid': '1234'}
        True


    Implementing structured logging

        >>> import pygogo as gogo
        >>>
        >>> s = StringIO()
        >>> hdlr = gogo.handlers.fileobj_hdlr(s)
        >>> formatter = gogo.formatters.structured_formatter
        >>> kwargs = {
        ...     'low_level': 'info', 'low_formatter': formatter,
        ...     'low_hdlr': hdlr}
        >>> logger = gogo.Gogo('ex.one', **kwargs).logger
        >>> extra = {'set_value': set([1, 2, 3]), 'snowman': '☃'}
        >>> logger.info('log message', extra=extra)
        >>> result = loads(s.getvalue())
        >>> keys = sorted(result.keys())
        >>> keys == [
        ...     'level', 'message', 'msecs', 'name', 'set_value', 'snowman',
        ...     'time']
        True
        >>> blacklist = {'snowman', 'msecs', 'time'}
        >>> [result[k] for k in keys if k not in blacklist] == [
        ...     'INFO', 'log message', 'ex.one.base', [1, 2, 3]]
        True


    Multiple handlers and formatters

        >>> import logging
        >>> import pygogo as gogo
        >>>
        >>> log_format = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
        >>> formatter = logging.Formatter(log_format, datefmt='%Y')
        >>>
        >>> logger = gogo.Gogo(
        ...     'ex.two',
        ...     low_hdlr=gogo.handlers.file_hdlr('example2.log', mode='w'),
        ...     low_formatter=formatter,
        ...     high_level='error',
        ...     high_formatter=formatter).logger
        >>>
        >>> logger.debug('debug message')
        >>> logger.info('info message')
        >>> logger.warning('warn message')
        >>> logger.error('error message')  # doctest: +ELLIPSIS
        20... - ex.two.base - ERROR - error message
        >>> logger.critical('critical message')  # doctest: +ELLIPSIS
        20... - ex.two.base - CRITICAL - critical message

        >>> with open('example2.log', encoding='utf-8') as f:
        ...     [line.strip()[5:] for line in f] == [
        ...         '- ex.two.base - DEBUG - debug message',
        ...         '- ex.two.base - INFO - info message',
        ...         '- ex.two.base - WARNING - warn message',
        ...         '- ex.two.base - ERROR - error message',
        ...         '- ex.two.base - CRITICAL - critical message']
        True


    Logging to multiple destinations

        >>> import pygogo as gogo
        >>>
        >>> going = gogo.Gogo(
        ...     'ex.three',
        ...     low_hdlr=gogo.handlers.file_hdlr('example1.log', mode='w'),
        ...     low_formatter=gogo.formatters.fixed_formatter,
        ...     high_level='info',
        ...     high_formatter=gogo.formatters.console_formatter)
        >>>
        >>> logger1 = going.get_logger('area1')
        >>> logger2 = going.get_logger('area2')
        >>> root = going.logger
        >>>
        >>> root.info('Jackdaws love my big sphinx.')
        ex.three.base: INFO     Jackdaws love my big sphinx.
        >>> logger1.debug('Quick zephyrs blow, daft Jim.')
        >>> logger1.info('How daft jumping zebras vex.')
        ex.three.area1: INFO     How daft jumping zebras vex.
        >>> logger2.warning('Jail zesty vixen who pay.')
        ex.three.area2: WARNING  Jail zesty vixen who pay.
        >>> logger2.error('The five boxing wizards jump.')
        ex.three.area2: ERROR    The five boxing wizards jump.
        >>>
        >>> with open('example1.log', encoding='utf-8') as f:
        ...     [line.strip()[24:] for line in f] == [
        ...         'ex.three.base INFO     Jackdaws love my big sphinx.',
        ...         'ex.three.area1 DEBUG    Quick zephyrs blow, daft Jim.',
        ...         'ex.three.area1 INFO     How daft jumping zebras vex.',
        ...         'ex.three.area2 WARNING  Jail zesty vixen who pay.',
        ...         'ex.three.area2 ERROR    The five boxing wizards jump.']
        True


    Reset stderr so logs aren't printed twice

        >>> sys.stderr = sys.__stderr__


    Using Filters to impart contextual information

        >>> import logging
        >>> import pygogo as gogo
        >>>
        >>> levels = ('DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL')
        >>> log_frmt = (
        ...     '%(asctime)-4s %(name)-5s %(levelname)-8s IP: %(ip)-15s User: '
        ...     '%(user)-8s %(message)s')
        >>>
        >>> formatter = logging.Formatter(log_frmt, datefmt='%Y')
        >>> going = gogo.Gogo('a', low_formatter=formatter)
        >>> a1 = going.get_logger('b.c', ip='123.231.231.123', user='fred')
        >>> a2 = going.get_logger('e.f', ip='192.168.0.1', user='sheila')
        >>>
        >>> a1.debug('A debug message')  # doctest: +ELLIPSIS
        20... a.b.c DEBUG    IP: 123.231.231.123 User: fred     A debug message
        >>> a1.info('An info %s', 'message')  # doctest: +ELLIPSIS
        20... a.b.c INFO     IP: 123.231.231.123 User: fred     An info message
        >>>
        >>> for level in [getattr(logging, l) for l in levels]:
        ...     name = logging.getLevelName(level)
        ...     pronoun = 'AN' if name[0] in 'AEIOU' else 'A'
        ...     a2.log(level, '%s %s msg', pronoun, name)  # doctest: +ELLIPSIS
        20... a.e.f DEBUG    IP: 192.168.0.1     User: sheila   A DEBUG msg
        20... a.e.f INFO     IP: 192.168.0.1     User: sheila   AN INFO msg
        20... a.e.f WARNING  IP: 192.168.0.1     User: sheila   A WARNING msg
        20... a.e.f ERROR    IP: 192.168.0.1     User: sheila   AN ERROR msg
        20... a.e.f CRITICAL IP: 192.168.0.1     User: sheila   A CRITICAL msg
"""

from __future__ import (
    absolute_import, division, print_function, unicode_literals)

from builtins import *



# -*- coding: utf-8 -*-
# vim: sw=4:ts=4:expandtab

"""
tests
~~~~~

Provides application unit tests
"""

from __future__ import (
    absolute_import, division, print_function, unicode_literals)

import unittest
import re

from builtins import *

import pygogo as gogo

module_logger = gogo.Gogo(__name__).logger
initialized = False


def setup_package():
    """database context creation"""
    global initialized
    initialized = True
    module_logger.debug('Package Setup\n')


def teardown_package():
    """database context removal"""
    global initialized
    initialized = False
    module_logger.debug('Package Teardown\n')


class BaseTest(unittest.TestCase):
    def runTest(self, *args, **kwargs):
        pass

    def assertEqualEllipsis(self, expected, actual, marker='...', msg=None):
        """Checks whether actual is equal to expected while ignoring ellipsis
        content.

        # https://gist.github.com/harobed/5845674

        Args:
            expected (scalar): The expected value
            actual (scalar): The actual value

        Example:
            >>> BaseTest().assertEqualEllipsis('foo...bar', 'foo123bar')
        """
        if marker not in expected:
            self.assertEqual(expected, actual, msg)

        replaced = re.escape(expected).replace(re.escape(marker), '(.*?)')

        if re.match(replaced, actual, re.M | re.S) is None:
            self.assertMultiLineEqual(expected, actual, msg)

    def assertIsSubset(self, expected, actual):
        """Checks whether actual is a superset of expected.

        Args:
            expected (scalar): The expected value
            actual (scalar): The actual value

        Example:
            >>> BaseTest().assertIsSubset([1,2], range(5))
        """
        self.assertTrue(set(expected).issubset(actual))

    def assertIsNotSubset(self, expected, actual):
        """Checks whether actual is a superset of expected.

        Args:
            expected (scalar): The expected value
            actual (scalar): The actual value

        Example:
            >>> BaseTest().assertIsNotSubset([11,12], range(5))
        """
        self.assertFalse(set(expected).issubset(actual))



# -*- coding: utf-8 -*-
# vim: sw=4:ts=4:expandtab

"""
pygogo.formatters
~~~~~~~~~~~~~~~~~

Log formatters

Examples:
    Add a console formatter::

        >>> import sys

        >>> logger = logging.getLogger('console_logger')
        >>> hdlr = logging.StreamHandler(sys.stdout)
        >>> hdlr.setFormatter(console_formatter)
        >>> logger.addHandler(hdlr)
        >>> logger.info('hello world')
        console_logger: INFO     hello world

    Add a structured formatter::

        >>> from io import StringIO
        >>> from json import loads

        >>> s = StringIO()
        >>> logger = logging.getLogger('structured_logger')
        >>> hdlr = logging.StreamHandler(s)
        >>> hdlr.setFormatter(structured_formatter)
        >>> extra = {'key': 'value'}
        >>> logger.addHandler(hdlr)
        >>> logger.info('hello world', extra=extra)
        >>> result = loads(s.getvalue())
        >>> keys = sorted(result.keys())
        >>> keys == ['key', 'level', 'message', 'msecs', 'name', 'time']
        True
        >>> [result[k] for k in keys if k not in {'msecs', 'time'}] == [
        ...     'value', 'INFO', 'hello world', 'structured_logger']
        True

Attributes:
    BASIC_FORMAT (str): A basic format

    CONSOLE_FORMAT (str): A format for displaying in a console

    FIXED_FORMAT (str): A fixed width format

    CSV_FORMAT (str): A csv format

    JSON_FORMAT (str): A json format

    DATEFMT (str): Standard date format
"""

from __future__ import (
    absolute_import, division, print_function, unicode_literals)

import logging
import sys
import traceback
import itertools as it

from builtins import *
from .utils import CustomEncoder

BASIC_FORMAT = '%(message)s'
BOM_FORMAT = '\ufeff%(message)s'
CONSOLE_FORMAT = '%(name)-12s: %(levelname)-8s %(message)s'
FIXED_FORMAT = '%(asctime)s.%(msecs)-3d %(name)-12s %(levelname)-8s %(message)s'
CSV_FORMAT = '%(asctime)s.%(msecs)d,%(name)s,%(levelname)s,"%(message)s"'
JSON_FORMAT = (
    '{"time": "%(asctime)s.%(msecs)d", "name": "%(name)s", "level":'
    ' "%(levelname)s", "message": "%(message)s"}')

DATEFMT = '%Y-%m-%d %H:%M:%S'

module_hdlr = logging.StreamHandler(sys.stdout)
module_logger = logging.getLogger(__name__)
module_logger.addHandler(module_hdlr)


class StructuredFormatter(logging.Formatter):
    """A logging formatter that creates a json string from log details

    Args:
        fmt (string): Log message format.

        datefmt (dict): Log date format.

    Returns:
        New instance of :class:`StructuredFormatter`

    Examples:
        >>> from io import StringIO
        >>> from json import loads

        >>> s = StringIO()
        >>> logger = logging.getLogger()
        >>> formatter = StructuredFormatter(BASIC_FORMAT, datefmt=DATEFMT)
        >>> hdlr = logging.StreamHandler(s)
        >>> hdlr.setFormatter(formatter)
        >>> logger.addHandler(hdlr)
        >>> logger.info('hello world')
        >>> result = loads(s.getvalue())
        >>> keys = sorted(result.keys())
        >>> keys == ['level', 'message', 'msecs', 'name', 'time']
        True
        >>> [result[k] for k in keys if k not in {'msecs', 'time'}] == [
        ...     'INFO', 'hello world', 'root']
        True
    """
    def __init__(self, fmt=None, datefmt=None):
        """Initialization method.

        Args:
            fmt (string): Log message format.

            datefmt (dict): Log date format.

        Returns:
            New instance of :class:`StructuredFormatter`

        Examples:
            >>> StructuredFormatter('name')  # doctest: +ELLIPSIS
            <pygogo.formatters.StructuredFormatter object at 0x...>
        """
        empty_record = logging.makeLogRecord({})
        filterer = lambda k: k not in empty_record.__dict__ and k != 'asctime'
        self.filterer = filterer
        super(StructuredFormatter, self).__init__(fmt, datefmt)

    def format(self, record):
        """ Formats a record as a dict string

        Args:
            record (object): The event to format.

        Returns:
            str: The formatted content

        Examples:
            >>> from json import loads

            >>> formatter = StructuredFormatter(BASIC_FORMAT, datefmt='%Y')
            >>> logger = logging.getLogger()
            >>> args = (logging.INFO, '.', 0, 'hello world', [], None)
            >>> record = logger.makeRecord('root', *args)
            >>> result = loads(formatter.format(record))
            >>> keys = sorted(result.keys())
            >>> keys == ['level', 'message', 'msecs', 'name', 'time']
            True
            >>> [result[k] for k in keys if k not in {'msecs', 'time'}] == [
            ...     'INFO', 'hello world', 'root']
            True
        """
        extra = {
            'message': record.getMessage(),
            'time': self.formatTime(record, self.datefmt),
            'msecs': record.msecs,
            'name': record.name,
            'level': record.levelname}

        keys = filter(self.filterer, record.__dict__)
        extra.update({k: record.__dict__[k] for k in keys})
        extra.pop('asctime', None)
        return str(CustomEncoder().encode(extra))

    def formatException(self, exc_info):
        """Formats an exception as a dict string

        Args:
            exc_info (tuple[type, value, traceback]): Exception tuple as
                returned by `sys.exc_info()`

        Returns:
            str: The formatted exception

        Examples:
            >>> from json import loads

            >>> formatter = StructuredFormatter(BASIC_FORMAT)
            >>> try:
            ...     1 / 0
            ... except:
            ...     result = loads(formatter.formatException(sys.exc_info()))
            >>> keys = sorted(result.keys())
            >>> keys == [
            ...     'filename', 'function', 'lineno', 'text', 'type', 'value']
            True
            >>> [result[k] for k in keys if k not in {'filename', 'type'}] == [
            ...     '<module>', 2, '1 / 0', 'division by zero']
            True
            >>> result['type'][-17:] == 'ZeroDivisionError'
            True
        """
        keys = ['type', 'value', 'filename', 'lineno', 'function', 'text']
        type_, value, trcbk = exc_info
        stype = str(type_).replace('type', '').strip(" '<>")
        values = it.chain([stype, value], *traceback.extract_tb(trcbk))
        return str(CustomEncoder().encode(dict(zip(keys, values))))

basic_formatter = logging.Formatter(BASIC_FORMAT)
bom_formatter = logging.Formatter(BOM_FORMAT)
console_formatter = logging.Formatter(CONSOLE_FORMAT)
fixed_formatter = logging.Formatter(FIXED_FORMAT, datefmt=DATEFMT)
csv_formatter = logging.Formatter(CSV_FORMAT, datefmt=DATEFMT)
json_formatter = logging.Formatter(JSON_FORMAT, datefmt=DATEFMT)
structured_formatter = StructuredFormatter(BASIC_FORMAT, datefmt=DATEFMT)



# -*- coding: utf-8 -*-
# vim: sw=4:ts=4:expandtab

"""
pygogo.utils
~~~~~~~~~~~~

Misc classes and functions that don't warrant their own module

Examples:
    basic usage::

        >>> CustomEncoder().encode(range(5))
        '[0, 1, 2, 3, 4]'

"""

from __future__ import (
    absolute_import, division, print_function, unicode_literals)

import logging
import sys

from json import JSONEncoder
from builtins import *

module_hdlr = logging.StreamHandler(sys.stdout)
module_logger = logging.getLogger(__name__)
module_logger.addHandler(module_hdlr)


class CustomEncoder(JSONEncoder):
    """A unicode aware JSON encoder that can handle iterators, dates, and times

    Examples:
        >>> CustomEncoder().encode(range(5))
        '[0, 1, 2, 3, 4]'
        >>> from json import dumps
        >>> dumps(range(5), cls=CustomEncoder)
        '[0, 1, 2, 3, 4]'
    """
    def default(self, obj):
        """ Encodes a given object

        Args:
            obj (scalar): The object to encode.

        Returns:
            The encoded object

        Examples:
            >>> CustomEncoder().default(range(5))
            [0, 1, 2, 3, 4]
        """
        if hasattr(obj, 'real'):
            encoded = float(obj)
        elif hasattr(obj, 'union'):
            encoded = tuple(obj)
        elif set(['next', 'union', '__iter__']).intersection(dir(obj)):
            encoded = list(obj)
        else:
            encoded = str(obj)

        return encoded


class StructuredMessage(object):
    """Converts a message and kwargs to a json string

    Attributes:
        kwargs (dict): Keyword arguments passed to
            :class:`~pygogo.utils.CustomEncoder`.

    Args:
        message (string): The message to log.

        kwargs (dict): Keyword arguments passed to
            :class:`~pygogo.utils.CustomEncoder`.

    Returns:
        New instance of :class:`StructuredMessage`

    See also:
        :class:`pygogo.utils.StructuredAdapter`

    Examples:
        >>> from json import loads

        >>> msg = StructuredMessage('hello world', key='value')
        >>> loads(str(msg)) == {'message': 'hello world', 'key': 'value'}
        True
    """
    def __init__(self, message=None, **kwargs):
        """Initialization method.

        Args:
            message (string): The message to log.

            kwargs (dict): Keyword arguments passed to
                :class:`~pygogo.utils.CustomEncoder`.

        Returns:
            New instance of :class:`StructuredMessage`

        Examples:
            >>> StructuredMessage('message')  # doctest: +ELLIPSIS
            <pygogo.utils.StructuredMessage object at 0x...>
        """
        kwargs['message'] = message
        self.kwargs = kwargs

    def __str__(self):
        """ String method

        Returns:
            str: The encoded object

        Examples
            >>> from json import loads

            >>> msg = str(StructuredMessage('hello world', key='value'))
            >>> loads(msg) == {'message': 'hello world', 'key': 'value'}
            True

        """
        return str(CustomEncoder().encode(self.kwargs))


class StructuredAdapter(logging.LoggerAdapter):
    """A logging adapter that creates a json string from a log message and the
    `extra` kwarg

    See also:
        :class:`pygogo.utils.StructuredMessage`

        :meth:`pygogo.Gogo.get_structured_logger`

    Examples:
        >>> from io import StringIO
        >>> from json import loads

        >>> s = StringIO()
        >>> logger = logging.getLogger()
        >>> hdlr = logging.StreamHandler(s)
        >>> logger.addHandler(hdlr)
        >>> structured_logger = StructuredAdapter(logger, {'all': True})
        >>> structured_logger.debug('hello', extra={'key': u'value'})
        >>> loads(s.getvalue()) == {
        ...     'all': True, 'message': 'hello', 'key': 'value'}
        True
    """
    def process(self, msg, kwargs):
        """ Modifies the message and/or keyword arguments passed to a logging
        call in order to insert contextual information.

        Args:
            msg (str): The message to log.
            kwargs (dict):

        Returns:
            Tuple of (:class:`~pygogo.utils.StructuredMessage`, modified kwargs)

        Examples:
            >>> from json import loads

            >>> logger = logging.getLogger()
            >>> structured_logger = StructuredAdapter(logger, {'all': True})
            >>> extra = {'key': 'value'}
            >>> m, k = structured_logger.process('message', {'extra': extra})
            >>> loads(m) == {'all': True, 'message': 'message', 'key': 'value'}
            True
            >>> k == {'extra': {'all': True, 'key': 'value'}}
            True
        """
        extra = kwargs.get('extra', {})
        extra.update(self.extra)
        kwargs['extra'] = extra
        return str(StructuredMessage(msg, **extra)), kwargs


class LogFilter(logging.Filter):
    """Filters log messages depending on level

    Attributes:
        level (int): The logging level.

            +-------------------------+-------+
            | logging level attribute | value |
            +=========================+=======+
            | CRITICAL                | 50    |
            +-------------------------+-------+
            | ERROR                   | 40    |
            +-------------------------+-------+
            | WARNING                 | 30    |
            +-------------------------+-------+
            | INFO                    | 20    |
            +-------------------------+-------+
            | DEBUG                   | 10    |
            +-------------------------+-------+
            | NOTSET                  |  0    |
            +-------------------------+-------+

    Args:
        level (int): The logging level.

    Returns:
        New instance of :class:`LogFilter`

    See also:
        :meth:`pygogo.Gogo.update_hdlr`
    """
    def __init__(self, level):
        """Initialization method.

        Args:
            level (int): The logging level.

        Returns:
            New instance of :class:`LogFilter`

        Examples:
            >>> LogFilter(40)  # doctest: +ELLIPSIS
            <pygogo.utils.LogFilter object at 0x...>
        """
        self.high_level = level

    def filter(self, record):
        """Determines whether or a not a message should be logged.

        Args:
            record (obj): The event to (potentially) log

        Returns:
            bool: True if the event level is lower than self.high_level

        Examples:
            >>> attrs = {'levelno': logging.INFO}
            >>> record = logging.makeLogRecord(attrs)
            >>> LogFilter(40).filter(record)
            True
        """
        return record.levelno < self.high_level


def get_structured_filter(name='', **kwargs):
    """Returns a structured filter that injects contextual information into
    log records.

    Args:
        kwargs (dict): The contextual information you wish to inject

    See also:
        :meth:`pygogo.Gogo.update_hdlr`

    Returns:
        New instance of :class:`pygogo.utils.StructuredFilter`

    Examples:
        >>> structured_filter = get_structured_filter(user='fred')
        >>> structured_filter  # doctest: +ELLIPSIS
        <pygogo.utils...StructuredFilter object at 0x...>
        >>>
        >>> logger = logging.getLogger('structured_filter')
        >>> hdlr = logging.StreamHandler(sys.stdout)
        >>> formatter = logging.Formatter('User %(user)s said, "%(message)s".')
        >>> hdlr.setFormatter(formatter)
        >>> logger.addFilter(structured_filter)
        >>> logger.addHandler(hdlr)
        >>> logger.debug('A debug message')
        User fred said, "A debug message".

    """
    class StructuredFilter(logging.Filter):
        """
        Injects contextual information into log records.
        """
        def filter(self, record):
            """Adds contextual information to a log record

            Args:
                record (obj): The event to contextualize

            Returns:
                bool: True
            """
            for k, v in kwargs.items():
                setattr(record, k, v)

            return True

    return StructuredFilter(name)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim: sw=4:ts=4:expandtab

""" A Python logging library with super powers """

from __future__ import (
    absolute_import, division, print_function, unicode_literals)

import sys
import itertools as it

from os import getcwd, path as p
from argparse import RawTextHelpFormatter, ArgumentParser
from builtins import *

import pygogo as gogo

HDLRS_FULL = tuple(h for h in dir(gogo.handlers) if h.endswith('hdlr'))
HDLRS = tuple(h[:-5] for h in HDLRS_FULL)
LEVELS = ('critical', 'error', 'warning', 'info', 'debug')
FRMTRS_FULL = tuple(f for f in dir(gogo.formatters) if f.endswith('formatter'))
FORMATS = tuple(f[:-10] for f in FRMTRS_FULL)
CURDIR = p.basename(getcwd())
LOGFILE = '%s.log' % CURDIR

parser = ArgumentParser(
    description='description: Logs a given message', prog='gogo',
    usage='%(prog)s [options] <message>', formatter_class=RawTextHelpFormatter)

parser.add_argument(
    dest='message', nargs='?', default=sys.stdin,
    help='The message to log (default: reads from stdin).')

parser.add_argument(
    '-l', '--level', metavar='LEVEL', choices=LEVELS, default='info',
    help=(
        "The level to log the message (default: info).\n"
        "Must be one of: %s,\n%s.\n\n" % (
            ', '.join(LEVELS[:4]), ', '.join(LEVELS[4:]))))

parser.add_argument(
    '-n', '--name', default=CURDIR,
    help="The logger name (default: %s).\n\n" % CURDIR)

parser.add_argument(
    '-D', '--high-hdlr', metavar='HANDLER', choices=HDLRS, default='stderr',
    help=(
        "The high pass log handler (default: stderr).\n"
        "Must be one of: %s,\n%s.\n\n" % (
            ', '.join(HDLRS[:4]), ', '.join(HDLRS[4:]))))

parser.add_argument(
    '-d', '--low-hdlr', metavar='HANDLER', choices=HDLRS, default='stdout',
    help=(
        "The low pass log handler (default: stdout).\n"
        "Must be one of: %s,\n%s.\n\n" % (
            ', '.join(HDLRS[:4]), ', '.join(HDLRS[4:]))))

parser.add_argument(
    '-L', '--high-level', metavar='LEVEL', choices=LEVELS, default='warning',
    help=(
        "Min level to log to the high pass handler\n"
        "(default: warning). Must be one of: %s,\n%s.\n\n" % (
            ', '.join(LEVELS[:1]), ', '.join(LEVELS[1:]))))

parser.add_argument(
    '-e', '--low-level', metavar='LEVEL', choices=LEVELS, default='debug',
    help=(
        "Min level to log to the low pass handler\n"
        "(default: debug). Must be one of: %s,\n%s.\n\n" % (
            ', '.join(LEVELS[:1]), ', '.join(LEVELS[1:]))))

parser.add_argument(
    '-F', '--high-format', metavar='FORMAT', choices=FORMATS, default='basic',
    help=(
        "High pass handler log format (default: basic)."
        "\nMust be one of: %s,\n%s.\n\n" % (
            ', '.join(FORMATS[:4]), ', '.join(FORMATS[4:]))))

parser.add_argument(
    '-o', '--low-format', metavar='FORMAT', choices=FORMATS, default='basic',
    help=(
        "Low pass handler log format (default: basic)."
        "\nMust be one of: %s,\n%s.\n\n" % (
            ', '.join(FORMATS[:4]), ', '.join(FORMATS[4:]))))

parser.add_argument(
    '-m', '--monolog', action='store_true', default=False,
    help="Log high level events only to high pass handler.\n\n")

parser.add_argument(
    '-f', '--filename', action='append', default=[LOGFILE],
    help=(
        "The filename to log to  (default: %s).\nUsed in the following "
        "handlers: file.\n\n") % LOGFILE)

parser.add_argument(
    '-s', '--subject', default=["You've got mail"], action='append',
    help=(
        "The log subject (default: You've got mail)."
        "\nUsed in the following handlers: email.\n\n"))

parser.add_argument(
    '-u', '--url', action='append', default=[''],
    help="The log url. Required for the following handlers:\nwebhook.\n\n")

parser.add_argument(
    '-H', '--host', default=['localhost'], action='append',
    help=(
        "The host (default: localhost).\nUsed in the following handlers: "
        "socket and syslog.\n\n"))

parser.add_argument(
    '-p', '--port', metavar='NUM', type=int, action='append', default=[''],
    help=(
        "The port number (default: Python logging default).\nUsed in the "
        "following handlers: socket and syslog.\n\n"))

parser.add_argument(
    '-t', '--tcp', action='count', default=0, help=(
        "Use TCP instead of UDP.\nUsed in the following handlers: socket and "
        "syslog.\n\n"))

parser.add_argument(
    '-g', '--get', action='count', default=0, help=(
        "Use a GET request instead of POST.\nUsed in the following handlers: "
        "webhook.\n\n"))

parser.add_argument(
    '-v', '--version', help="Show version and exit.", action='store_true',
    default=False)

parser.add_argument(
    '-V', '--verbose', help='Increase output verbosity.', action='store_true',
    default=False)


def run():
    """CLI runner
    """
    args = parser.parse_args()
    gogo_logger = gogo.Gogo(__name__, verbose=args.verbose).get_logger('run')

    if args.version:
        gogo_logger.info('gogo v%s' % gogo.__version__)
        exit(0)

    counted = {'get', 'tcp'}
    appended = {'filename', 'subject', 'url', 'host', 'port'}
    items = args._get_kwargs()
    counted_args = [i for i in items if i[0] in counted]
    appended_args = [i for i in items if i[0] in appended]

    high_appended_args = [(k, v[0]) for k, v in appended_args]
    high_counted_args = [(k, v > 0) for k, v in counted_args]
    high_kwargs = dict(it.chain(high_appended_args, high_counted_args))

    low_appended_args = [(k, v[-1]) for k, v in appended_args]
    low_counted_args = [(k, v > 1) for k, v in counted_args]
    low_kwargs = dict(it.chain(low_appended_args, low_counted_args))

    high_hdlr = getattr(gogo.handlers, '%s_hdlr' % args.high_hdlr)
    low_hdlr = getattr(gogo.handlers, '%s_hdlr' % args.low_hdlr)
    high_format = getattr(gogo.formatters, '%s_formatter' % args.high_format)
    low_format = getattr(gogo.formatters, '%s_formatter' % args.low_format)

    nkwargs = {
        'verbose': args.verbose,
        'high_level': args.high_level.upper(),
        'low_level': args.low_level.upper(),
        'high_formatter': high_format,
        'low_formatter': low_format,
        'monolog': args.monolog,
        'high_hdlr': high_hdlr(**high_kwargs),
        'low_hdlr': low_hdlr(**low_kwargs)}

    logger = gogo.Gogo(args.name, **nkwargs).get_logger('runner')

    try:
        message = args.message.read()
    except AttributeError:
        message = args.message

    getattr(logger, args.level)(message)
    exit(0)

if __name__ == '__main__':
    run()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim: sw=4:ts=4:expandtab

"""
tests.test
~~~~~~~~~~

Provides scripttests pygogo CLI functionality.
"""

from __future__ import (
    absolute_import, division, print_function, unicode_literals)

import sys

sys.path.append('../pygogo')
import pygogo as gogo                       # noqa

from difflib import unified_diff            # noqa
from os import path as p                    # noqa
from io import StringIO                     # noqa
from timeit import default_timer as timer   # noqa

from builtins import *                      # noqa
from scripttest import TestFileEnvironment  # noqa


def main(script, tests, verbose=False, stop=True):
    """ Main method

    Returns 0 on success, 1 on failure
    """
    failures = 0
    logger = gogo.Gogo(__name__, verbose=verbose).logger
    short_script = p.basename(script)
    env = TestFileEnvironment('.scripttest')

    start = timer()

    for pos, test in enumerate(tests):
        num = pos + 1
        opts, arguments, expected = test
        joined_opts = ' '.join(opts) if opts else ''
        joined_args = '"%s"' % '" "'.join(arguments) if arguments else ''
        command = "%s %s %s" % (script, joined_opts, joined_args)
        short_command = "%s %s %s" % (short_script, joined_opts, joined_args)
        result = env.run(command, cwd=p.abspath(p.dirname(p.dirname(__file__))))
        output = result.stdout

        if isinstance(expected, bool):
            text = StringIO(output).read()
            outlines = [str(bool(text))]
            checklines = StringIO(str(expected)).readlines()
        elif p.isfile(expected):
            outlines = StringIO(output).readlines()

            with open(expected, encoding='utf-8') as f:
                checklines = f.readlines()
        else:
            outlines = StringIO(output).readlines()
            checklines = StringIO(expected).readlines()

        args = [checklines, outlines]
        kwargs = {'fromfile': 'expected', 'tofile': 'got'}
        diffs = ''.join(unified_diff(*args, **kwargs))
        passed = not diffs

        if not passed:
            failures += 1
            msg = "ERROR! Output from test #%i:\n  %s\n" % (num, short_command)
            msg += "doesn't match:\n  %s\n" % expected
            msg += diffs if diffs else ''
        else:
            logger.debug(output)
            msg = 'Scripttest #%i: %s ... ok' % (num, short_command)

        logger.info(msg)

        if stop and failures:
            break

    time = timer() - start
    logger.info('%s' % '-' * 70)
    end = 'FAILED (failures=%i)' % failures if failures else 'OK'
    logger.info('Ran %i scripttests in %0.3fs\n\n%s' % (num, time, end))
    sys.exit(failures)

if __name__ == '__main__':
    parent_dir = p.abspath(p.dirname(p.dirname(__file__)))
    script = p.join(parent_dir, 'bin', 'gogo')

    tests = [
        (['--help'], [], True),
        (['--version'], [], 'gogo v%s\n' % gogo.__version__),
        ([], ['hello world'], 'hello world\n'),
        (['-l debug'], ['hello world'], ''),
        (['-Vl debug'], ['hello world'], 'hello world\n'),
    ]

    main(script, tests)



# -*- coding: utf-8 -*-
# vim: sw=4:ts=4:expandtab

"""
pygogo.handlers
~~~~~~~~~~~~~~~

Log handlers

Examples:
    Add a stdout handler::

        >>> logger = logging.getLogger()
        >>> logger.addHandler(stdout_hdlr())
        >>> logger.info('hello world')
        hello world

Attributes:
    ENCODING (str): The module encoding
"""

from __future__ import (
    absolute_import, division, print_function, unicode_literals)

import sys
import logging
import socket

from os import environ
from logging import handlers as hdlrs
from builtins import *

try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse

ENCODING = 'utf-8'

module_hdlr = logging.StreamHandler(sys.stdout)
module_logger = logging.getLogger(__name__)
module_logger.addHandler(module_hdlr)


def stdout_hdlr(**kwargs):
    """A standard output log handler

    Returns:
        New instance of :class:`logging.StreamHandler`

    Examples:
        >>> stdout_hdlr()  # doctest: +ELLIPSIS
        <...StreamHandler...>
    """
    return logging.StreamHandler(sys.stdout)


def stderr_hdlr(**kwargs):
    """A standard error log handler

    Returns:
        New instance of :class:`logging.StreamHandler`

    Examples:
        >>> stderr_hdlr()  # doctest: +ELLIPSIS
        <...StreamHandler...>
    """
    return logging.StreamHandler(sys.stderr)


def fileobj_hdlr(f, **kwargs):
    """A file object log handler

    Args:
        f (obj): A file like object.

    Returns:
        New instance of :class:`logging.StreamHandler`

    Examples:
        >>> from io import StringIO
        >>> fileobj_hdlr(StringIO())  # doctest: +ELLIPSIS
        <...StreamHandler...>
    """
    return logging.StreamHandler(f)


def file_hdlr(filename, mode='a', encoding=ENCODING, delay=False, **kwargs):
    """A file log handler

    Args:
        filename (string): The logfile name.

        mode (string): The file open mode (default: a, i.e., append).

        encoding (string): The file encoding (default: the module ENCODING).

        delay (bool): Defer file opening until the first call to emit
            (default: False).

    Returns:
        New instance of :class:`logging.FileHandler`

    Examples:
        >>> from tempfile import NamedTemporaryFile
        >>> f = NamedTemporaryFile()
        >>> file_hdlr(f.name)  # doctest: +ELLIPSIS
        <...FileHandler...>
    """
    fkwargs = {'mode': mode, 'encoding': encoding, 'delay': delay}
    return logging.FileHandler(filename, **fkwargs)


def socket_hdlr(host='localhost', port=None, tcp=False, **kwargs):
    """A socket log handler

    Args:
        host (string): The host name (default: localhost).

        port (int): The port (default: `logging.handlers` default).

        tcp (bool): Create a TCP connection instead of UDP (default: False).

    Returns:
        New instance of either :class:`logging.handlers.DatagramHandler` or
            :class:`logging.handlers.SocketHandler`

    Examples:
        >>> socket_hdlr()  # doctest: +ELLIPSIS
        <...DatagramHandler...>
        >>> socket_hdlr(tcp=True)  # doctest: +ELLIPSIS
        <...SocketHandler...>
    """
    if tcp:
        def_port = hdlrs.DEFAULT_TCP_LOGGING_PORT
        handler = hdlrs.SocketHandler
    else:
        def_port = hdlrs.DEFAULT_UDP_LOGGING_PORT
        handler = hdlrs.DatagramHandler

    address = (host, port or def_port)
    return handler(*address)


def syslog_hdlr(host='localhost', port=None, tcp=False, **kwargs):
    """A syslog log handler

    Args:
        host (string): The host name (default: localhost). Set to None to use
            the platform dependent domain socket.

        port (int): The port (default: `logging.handlers` default).

        tcp (bool): Create a TCP connection instead of UDP (default: False).

    Returns:
        New instance of :class:`logging.handlers.SysLogHandler`

    Examples:
        >>> syslog_hdlr()  # doctest: +ELLIPSIS
        <...SysLogHandler...>
    """
    # http://stackoverflow.com/a/13874620/408556
    DEF_SOCKETS = {'linux2': '/dev/log', 'darwin': '/var/run/syslog'}

    if tcp:
        def_port = hdlrs.SYSLOG_TCP_PORT
        socktype = socket.SOCK_STREAM
    else:
        def_port = hdlrs.SYSLOG_UDP_PORT
        socktype = socket.SOCK_DGRAM

    if kwargs.get('address'):
        address = kwargs['address']
    elif host:
        address = (host, port or def_port)
    elif sys.platform in DEF_SOCKETS:
        address = DEF_SOCKETS[sys.platform]
    else:
        msg = 'Domain socket location for {} is not supported.'
        raise ValueError(msg.format(sys.platform))

    if kwargs.get('facility'):
        facility = kwargs['facility']
    elif kwargs.get('local_num') and 8 > kwargs['local_num'] >= 0:
        # http://unix.stackexchange.com/a/146993
        value = 'LOG_LOCAL{}'.format(kwargs['facility'])
        facility = getattr(hdlrs.SysLogHandler, value)
    else:
        facility = hdlrs.SysLogHandler.LOG_USER

    return hdlrs.SysLogHandler(address, facility=facility, socktype=socktype)


def buffered_hdlr(target=None, capacity=4096, level='error', **kwargs):
    """A memory buffered log handler

    Args:
        target (obj): The target logger handler (default stdout).

        capacity (int): The buffer size (default 4096).

        level (string): The min event level required to flush buffer
            (default: error).

    Returns:
        New instance of :class:`logging.handlers.MemoryHandler`

    Examples:
        >>> buffered_hdlr()  # doctest: +ELLIPSIS
        <...MemoryHandler...>
    """
    target = target or logging.StreamHandler(sys.stdout)
    return hdlrs.MemoryHandler(capacity, level.upper(), target)


def webhook_hdlr(url, **kwargs):
    """A web log handler

    Args:
        url (string): The logging endpoint.

    Kwargs:
        get (bool): Use a GET request instead of POST (default: False).

    Returns:
        New instance of :class:`logging.handlers.HTTPHandler`

    Examples:
        >>> webhook_hdlr('http://api.mysite.com/log')  # doctest: +ELLIPSIS
        <...HTTPHandler...>
    """
    parsed = urlparse(url)
    secure = parsed.scheme == 'https'
    method = 'GET' if kwargs.get('get') else 'POST'
    args = (parsed.netloc, parsed.path)

    try:
        hdlr = hdlrs.HTTPHandler(*args, method=method, secure=secure)
    except TypeError:
        hdlr = hdlrs.HTTPHandler(*args, method=method)

    return hdlr


def email_hdlr(subject=None, **kwargs):
    """An email log handler

    Args:
        subject (str): The email subject (default: You've got mail.).

        kwargs(dict): Keyword arguments.

    Kwargs:
        host (str): The email server host (default: localhost).

        port (str): The email sever port (default: None).

        sender (str): The email sender (default: the system username at gmail).

        recipients (List[str]): The email recipients (default: the system
            username at gmail).

        username (str): The email sever username (default: None).

        password (str): The email sever password (default: None).

    Returns:
        New instance of :class:`logging.handlers.SMTPHandler`

    Examples:
        >>> email_hdlr('hello world')  # doctest: +ELLIPSIS
        <...SMTPHandler...>
    """
    host = kwargs.get('host', 'localhost')
    port = kwargs.get('port')
    address = (host, port) if port else host
    sender = kwargs.get('sender', '%s@gmail.com' % environ.get('USER'))
    def_recipient = '%s@gmail.com' % environ.get('USER')
    recipients = kwargs.get('recipients', [def_recipient])
    subject = kwargs.get('subject', "You've got mail")
    username = kwargs.get('username')
    password = kwargs.get('password')

    args = (address, sender, recipients, subject)
    credentials = (username, password) if username or password else None
    return hdlrs.SMTPHandler(*args, credentials=credentials)



# -*- coding: utf-8 -*-
# vim: sw=4:ts=4:expandtab

"""
tests.test_main
~~~~~~~~~~~~~~~

Provides unit tests for the website.
"""

from __future__ import (
    absolute_import, division, print_function, unicode_literals)

import nose.tools as nt
import logging
import sys
import pygogo as gogo

from io import StringIO
from json import loads
from builtins import *
from . import BaseTest

module_logger = gogo.Gogo(__name__).logger
nt.assert_equal_ellipsis = BaseTest().assertEqualEllipsis
nt.assert_is_subset = BaseTest().assertIsSubset
nt.assert_is_not_subset = BaseTest().assertIsNotSubset


def setup_module():
    """site initialization"""
    global initialized
    initialized = True
    module_logger.debug('Main module setup\n')


class TestMain(BaseTest):
    """Main unit tests"""
    cls_initialized = False

    def setUp(self):
        nt.assert_false(self.cls_initialized)
        self.cls_initialized = True
        module_logger.debug('TestMain class setup\n')

    def tearDown(self):
        nt.ok_(self.cls_initialized)
        module_logger.debug('TestMain class teardown\n')

    def test_formatters(self):
        # basic
        sys.stderr = StringIO()
        logger = gogo.Gogo('stderr_if_gt_error', 'error').logger
        logger.warning('stdout')
        logger.error('stderr')

        # json_formatter
        formatter = gogo.formatters.json_formatter
        json_logger = gogo.Gogo('json', low_formatter=formatter).logger
        json_logger.debug('hello')

        # csv_formatter
        formatter = gogo.formatters.csv_formatter
        csv_logger = gogo.Gogo('csv', low_formatter=formatter).logger
        csv_logger.debug('hello')

        # console_formatter
        formatter = gogo.formatters.console_formatter
        console_lggr = gogo.Gogo('console', low_formatter=formatter).logger
        console_lggr.debug('hello')

        console_msg = (
            'stdout\nstderr\n{"time": "20...", "name": "json.base", "level": '
            '"DEBUG", "message": "hello"}\n20...,csv.base,DEBUG,"hello"\n'
            'console.base: DEBUG    hello')

        results = sys.stdout.getvalue().strip()
        nt.assert_equal_ellipsis(console_msg, results)
        nt.assert_equal('stderr', sys.stderr.getvalue().strip())

    def test_handlers(self):
        f = StringIO()
        hdlr = gogo.handlers.fileobj_hdlr(f)
        lggr = gogo.Gogo('test_handlers', high_hdlr=hdlr).logger

        msg1 = 'stdout hdlr only'
        lggr.debug(msg1)
        f.seek(0)
        nt.assert_equal(msg1, sys.stdout.getvalue().strip())
        nt.assert_false(f.read())

        msg2 = 'both hdlrs'
        lggr.error(msg2)
        f.seek(0)
        nt.assert_equal('%s\n%s' % (msg1, msg2), sys.stdout.getvalue().strip())
        nt.assert_equal(f.read().strip(), msg2)

    def test_multiple_loggers(self):
        f = StringIO()

        going = gogo.Gogo(
            'myapp',
            low_hdlr=gogo.handlers.fileobj_hdlr(f),
            low_formatter=gogo.formatters.fixed_formatter,
            high_hdlr=gogo.handlers.stdout_hdlr(),
            high_level='info',
            high_formatter=gogo.formatters.console_formatter)

        root = going.logger
        logger1 = going.get_logger('area1')
        logger2 = going.get_logger('area2')

        root.info('Jackdaws love my big sphinx of quartz.')
        logger1.debug('Quick zephyrs blow, vexing daft Jim.')
        logger1.info('How quickly daft jumping zebras vex.')
        logger2.warning('Jail zesty vixen who grabbed pay.')
        logger2.error('The five boxing wizards jump quickly.')

        console_msg = (
            "myapp.base  : INFO     Jackdaws love my big sphinx of quartz."
            "\nmyapp.area1 : INFO     How quickly daft jumping zebras vex."
            "\nmyapp.area2 : WARNING  Jail zesty vixen who grabbed pay."
            "\nmyapp.area2 : ERROR    The five boxing wizards jump quickly.")

        file_msg = [
            "myapp.base   INFO     Jackdaws love my big sphinx of quartz.\n",
            "myapp.area1  DEBUG    Quick zephyrs blow, vexing daft Jim.\n",
            "myapp.area1  INFO     How quickly daft jumping zebras vex.\n",
            "myapp.area2  WARNING  Jail zesty vixen who grabbed pay.\n",
            "myapp.area2  ERROR    The five boxing wizards jump quickly.\n"]

        f.seek(0)
        nt.assert_equal(console_msg, sys.stdout.getvalue().strip())
        nt.assert_equal([line[24:] for line in f], file_msg)

    def test_params_and_looping(self):
        levels = ('DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL')
        f = StringIO()
        going = gogo.Gogo(low_hdlr=gogo.handlers.fileobj_hdlr(f))
        logger1 = going.get_logger('area1')
        logger2 = gogo.Gogo().get_logger('area2')

        logger1_msg = 'A debug message\nAn info message'
        logger1.debug('A debug message')
        logger1.info('An info %s', 'message')
        f.seek(0)
        nt.assert_equal(f.read().strip(), logger1_msg)
        logger2_msg = ''

        for level in [getattr(logging, l) for l in levels]:
            name = logging.getLevelName(level)
            logger2_msg += '%s message\n' % name
            logger2.log(level, '%s %s', name, 'message')
            # TODO: lookup yielding from a nose test

        nt.assert_equal(logger2_msg.strip(), sys.stdout.getvalue().strip())

    def test_debugging(self):
        console_msg = (
            'This should log\nSo should this\nAnd this too\nThis should log '
            'also\nSo should this\nAnd this too\nBut this one should\nAnd this'
            ' one too')

        logger = gogo.Gogo('debug.none').logger
        logger.debug('This should log')
        logger.info('So should this')
        logger.warning('And this too')

        logger = gogo.Gogo('debug.on', verbose=True).logger
        logger.debug('This should log also')
        logger.info('So should this')
        logger.warning('And this too')

        logger = gogo.Gogo('debug.off', verbose=False).logger
        logger.debug("This shouldn't log")
        logger.info('But this one should')
        logger.warning('And this one too')

        results = sys.stdout.getvalue().strip()
        nt.assert_in("This should log", results)
        nt.assert_not_in("This shouldn't log", results)
        nt.assert_equal(console_msg, results)

    def test_structured_formatter(self):
        console_msg = {
            'snowman': '\u2603', 'name': 'structured_formatter.base',
            'level': 'INFO', 'message': 'log message', 'time': '20...',
            'msecs': '...', 'set_value': [1, 2, 3]}

        log_format = gogo.formatters.BASIC_FORMAT
        skwargs = {'datefmt': '%Y'}
        formatter = gogo.formatters.StructuredFormatter(log_format, **skwargs)

        kwargs = {'low_level': 'info', 'low_formatter': formatter}
        logger = gogo.Gogo('structured_formatter', **kwargs).logger
        extra = {'set_value': set([1, 2, 3]), 'snowman': '\u2603'}
        logger.info('log message', extra=extra)
        result = loads(sys.stdout.getvalue())
        result['msecs'] = str(result['msecs'])
        keys = sorted(result.keys())
        nt.assert_equal(sorted(console_msg.keys()), keys)
        whitelist = {'msecs', 'time'}

        for k in keys:
            f = nt.assert_equal_ellipsis if k in whitelist else nt.assert_equal
            f(console_msg[k], result[k])

    def test_structured_logging(self):
        kwargs = {'persist': True}
        extra = {'additional': True}
        meta = set(['level', 'name', 'time'])

        # Basic structured logger
        logger0 = gogo.Gogo('logger0').get_structured_logger('base', **kwargs)

        # Structured formatter
        formatter = gogo.formatters.structured_formatter
        logger1 = gogo.Gogo('logger1', low_formatter=formatter).logger

        # JSON formatter
        formatter = gogo.formatters.json_formatter
        logger2 = gogo.Gogo('logger2', low_formatter=formatter).logger

        # Custom logger
        logfmt = (
            '{"time": "%(asctime)s.%(msecs)d", "name": "%(name)s", "level":'
            ' "%(levelname)s", "message": "%(message)s", '
            '"persist": "%(persist)s", "additional": "%(additional)s"}')

        fmtr = logging.Formatter(logfmt, datefmt=gogo.formatters.DATEFMT)
        logger3 = gogo.Gogo('logger3', low_formatter=fmtr).get_logger(**kwargs)

        # Now log some messages
        for logger in [logger0, logger1, logger2, logger3]:
            logger.debug('message', extra=extra)

        lines = sys.stdout.getvalue().strip().split('\n')
        results = [loads(l) for l in lines]

        # Assert the following loggers provide the log event meta data
        nt.assert_is_not_subset(meta, results[0])
        nt.assert_is_subset(meta, results[1])
        nt.assert_is_subset(meta, results[2])
        nt.assert_is_subset(meta, results[3])

        # Assert the following loggers provide the `extra` information
        nt.assert_in('additional', results[0])
        nt.assert_in('additional', results[1])
        nt.assert_not_in('additional', results[2])
        nt.assert_in('additional', results[3])

        # Assert the following loggers provide the `persist` information
        nt.assert_in('persist', results[0])
        nt.assert_in('persist', results[1])
        nt.assert_not_in('persist', results[2])
        nt.assert_in('persist', results[3])

        # Assert the following loggers provide the `msecs` in the time
        nt.assert_false(len(results[0].get('time', [])))
        nt.assert_false(len(results[1]['time'][20:]))
        nt.ok_(len(results[2]['time'][20:]))
        nt.ok_(len(results[3]['time'][20:]))

    def test_named_loggers(self):
        sys.stderr = sys.stdout
        logger1 = gogo.Gogo('named').logger
        logger2 = gogo.Gogo('named').logger
        nt.assert_equal(logger1, logger2)

        formatter = gogo.formatters.structured_formatter

        going = gogo.Gogo('named2', low_formatter=formatter)
        logger1 = going.get_logger('foo', test='foo')
        logger2 = going.get_logger('bar', test='bar')
        logger1.debug('message')
        logger2.debug('message')

        for h1, h2 in zip(logger1.handlers, logger2.handlers):
            nt.assert_not_equal(h1, h2)

            for f1, f2 in zip(h1.filters, h2.filters):
                nt.assert_not_equal(f1, f2)

        hdlr = gogo.handlers.stdout_hdlr()
        going = gogo.Gogo('named3', low_hdlr=hdlr, low_formatter=formatter)
        logger1 = going.get_logger('baz', test='baz')
        logger2 = going.get_logger('buzz', test='buzz')

        logger1.debug('message')
        logger2.debug('message')

        for h1, h2 in zip(logger1.handlers, logger2.handlers):
            nt.assert_not_equal(h1, h2)

            for f1, f2 in zip(h1.filters, h2.filters):
                nt.assert_not_equal(f1, f2)

        lines = sys.stdout.getvalue().strip().split('\n')
        nt.assert_not_equal(*(loads(l)['test'] for l in lines[0:2]))
        nt.assert_not_equal(*(loads(l)['test'] for l in lines[2:4]))



# -*- coding: utf-8 -*-
# vim: sw=4:ts=4:expandtab

"""
pygogo
~~~~~~

A Python logging library with super powers

Examples:
    basic usage::

        >>> logger = Gogo('basic').logger
        >>> logger.debug('hello world')
        hello world

    intermediate usage::

        >>> formatter = logging.Formatter('IP: %(ip)s - %(message)s')
        >>> kwargs = {'low_formatter': formatter}
        >>> logger = Gogo('intermediate', **kwargs).get_logger(ip='1.1.1.1')
        >>> logger.debug('hello world')
        IP: 1.1.1.1 - hello world

    advanced usage::

        >>> from io import StringIO
        >>> from json import loads
        >>>
        >>> high = StringIO()
        >>> low = StringIO()
        >>> kwargs = {
        ...     'monolog': True, 'high_hdlr': handlers.fileobj_hdlr(high),
        ...     'low_hdlr': handlers.fileobj_hdlr(low)}
        >>> logger = Gogo('adv', **kwargs).get_structured_logger(ip='1.1.1.1')
        >>> logger.debug('debug')
        >>> logger.error('error')
        >>> loads(low.getvalue()) == {'ip': '1.1.1.1', 'message': 'debug'}
        True
        >>> loads(high.getvalue()) == {'ip': '1.1.1.1', 'message': 'error'}
        True
"""

from __future__ import (
    absolute_import, division, print_function, unicode_literals)

import logging
import hashlib
import sys

from copy import copy
from builtins import *
from . import formatters, handlers, utils

__version__ = '0.12.0'

__all__ = ['formatters', 'handlers', 'utils']
__title__ = 'pygogo'
__author__ = 'Reuben Cummings'
__description__ = 'A Python logging library with super powers'
__email__ = 'reubano@gmail.com'
__license__ = 'MIT'
__copyright__ = 'Copyright 2015 Reuben Cummings'

module_hdlr = logging.StreamHandler(sys.stdout)
module_fltr = logging.Filter(name='%s.init' % __name__)

# prevent handler from logging `pygogo.main` events
module_hdlr.addFilter(module_fltr)
module_logger = logging.getLogger(__name__)
module_logger.addHandler(module_hdlr)


class Gogo(object):
    """A logging class that logs events via different handlers depending on the
    severity

    http://stackoverflow.com/a/28743317/408556

    Attributes:
        loggers (set[str]): Set of existing loggers

        monolog (bool): Log high level events only to high pass handler.

        name (string): The logger name.

        levels (dict): The min levels to log to handlers.

        handlers (dict): The high/low pass log handlers.

        formatters (dict): The high/low pass log formatter.

        +------------------------------------+-----------------+
        | messages level                     | message handler |
        +====================================+=================+
        | < levels['low']                    | none            |
        +------------------------------------+-----------------+
        | >= levels['low'], < levels['high'] | low handler     |
        +------------------------------------+-----------------+
        | >= levels['high']                  | both handlers * |
        +------------------------------------+-----------------+

        * This is the case when :attr:`monolog` is `False`. If :attr:`monolog`
          is True, then :attr:`handlers['high']` will be the only
          message handler

    Args:
        name (string): The logger name.

        high_level (string): The min level to log to high_hdlr
            (default: warning).

        low_level (string): The min level to log to low_hdlr
            (default: debug).

        kwargs (dict): Keyword arguments.

    Kwargs:
        high_hdlr (obj): The high pass log handler (a :class:`logging.handlers`
            instance, default: `stderr` StreamHandler).

        low_hdlr (obj): The low pass log handler (a :class:`logging.handlers`
            instance, default: `stdout` StreamHandler).

        high_formatter (obj): The high pass log formatter (a
            :class:`logging.Formatter` instance, default:
            :data:`pygogo.formatters.basic_formatter`).

        low_formatter (obj): The low pass log formatter (a
            :class:`logging.Formatter` instance, default:
            :data:`pygogo.formatters.basic_formatter`).

        verbose (bool): If False, set low level to `info`, if True, set low
            level to `debug`, overrides `levels['low']` if specified
            (default: None).

        monolog (bool): Log high level events only to high pass handler (
            default: False)

    Returns:
        New instance of :class:`pygogo.Gogo`

    Examples:
        >>> Gogo('name').logger.debug('message')
        message
    """

    def __init__(self, name='root', high_level=None, low_level=None, **kwargs):
        """Initialization method.

        Examples:
            >>> Gogo('name') # doctest: +ELLIPSIS
            <pygogo.Gogo object at 0x...>
        """
        verbose = kwargs.get('verbose')
        high_level = high_level or 'warning'

        if verbose is None:
            low_level = low_level or 'debug'
        elif verbose:
            low_level = 'debug'
        else:
            low_level = 'info'

        self.levels = {
            'high': getattr(logging, high_level.upper(), None),
            'low': getattr(logging, low_level.upper(), None)}

        if not isinstance(self.levels['high'], int):
            raise ValueError('Invalid high_level: %s' % self.levels['high'])
        elif not isinstance(self.levels['low'], int):
            raise ValueError('Invalid low_level: %s' % self.levels['low'])
        elif self.levels['high'] < self.levels['low']:
            raise ValueError('high_level must be >= low_level')

        self.loggers = set()
        self.name = name
        self.handlers = {
            'high': kwargs.get('high_hdlr', handlers.stderr_hdlr()),
            'low': kwargs.get('low_hdlr', handlers.stdout_hdlr())}

        self.formatters = {
            'high': kwargs.get('high_formatter', formatters.basic_formatter),
            'low': kwargs.get('low_formatter', formatters.basic_formatter)}

        self.monolog = kwargs.get('monolog')

    @property
    def logger(self):
        """The logger property.

        Returns:
            New instance of :class:`logging.Logger`

        Examples:
            >>> from io import StringIO

            >>> s = StringIO()
            >>> sys.stderr = s
            >>> logger = Gogo('default').logger
            >>> logger  # doctest: +ELLIPSIS
            <...Logger...>
            >>> logger.debug('stdout')
            stdout
            >>> kwargs = {'low_level': 'info', 'monolog': True}
            >>> logger = Gogo('ignore_if_lt_info', **kwargs).logger
            >>> logger.debug('ignored')
            >>> logger.info('stdout')
            stdout
            >>> logger.warning('stderr')
            >>> s.getvalue().strip() == 'stderr'
            True
        """
        return self.get_logger()

    def update_hdlr(self, hdlr, level, formatter=None, monolog=False, **kwargs):
        """Update a handler with a formatter, level, and filters.

        Args:
            hdlr (obj): A :class:`logging.handlers` instance.

            level (int): The min level to log to to `hdlr`.

            formatter (obj): The log formatter (a :class:`logging.Formatter`
                instance, default: :data:`pygogo.formatters.basic_formatter`)

            monolog (bool): Log high level events only to high pass handler (
                default: False)

            kwargs (dict): Keyword arguemnts passed to
                `pygogo.utils.get_structured_filter`

        See also:
            :meth:`pygogo.Gogo.get_logger`

            :meth:`pygogo.Gogo.get_structured_logger`

            :func:`pygogo.utils.LogFilter`

            :func:`pygogo.utils.get_structured_filter`

        Examples:
            >>> low_hdlr = logging.StreamHandler(sys.stdout)
            >>> going = Gogo(low_hdlr=low_hdlr, monolog=True)
            >>> hdlr = going.handlers['low']
            >>> [hdlr.formatter, hdlr.filters, hdlr.level]
            [None, [], 0]
            >>> fmtr = going.formatters['low']
            >>> kwargs = {'formatter': fmtr, 'monolog': going.monolog}
            >>> going.update_hdlr(hdlr, going.levels['low'], **kwargs)
            >>> [hdlr.formatter, hdlr.filters, hdlr.level]
            ...  # doctest: +ELLIPSIS
            [<...Formatter obj...>, [<pygogo.utils.LogFilter obj...>], 10]
        """
        hdlr.setLevel(level)

        if monolog:
            log_filter = utils.LogFilter(self.levels['high'])
            hdlr.addFilter(log_filter)

        if kwargs:
            structured_filter = utils.get_structured_filter(**kwargs)
            hdlr.addFilter(structured_filter)

        hdlr.setFormatter(formatter)

    def zip(self, *fmtrs):
        """Format high/low handler properties so that they can be conveniently
            passed to `update_hdlr`.

        Args:
            fmtrs (seq[obj]): A sequence of :class:`logging.Formatter`
                instances.

        See also:
            :meth:`pygogo.Gogo.update_hdlr`

            :meth:`pygogo.Gogo.get_logger`

            :meth:`pygogo.Gogo.get_structured_logger`

        Returns:
            New instance of :class:`logging.handlers`

        Examples:
            >>> hdlr = logging.StreamHandler(sys.stdout)
            >>> copy_hdlr(hdlr) # doctest: +ELLIPSIS
            <...StreamHandler...>
        """
        hdlrs = [self.handlers['high'], self.handlers['low']]
        levels = [self.levels['high'], self.levels['low']]
        monologs = [False, self.monolog]
        return zip(hdlrs, levels, fmtrs, monologs)

    def get_logger(self, name='base', **kwargs):
        """Retrieve a named logger.

        Args:
            name (string): The logger name.

        See also:
            :func:`pygogo.copy_hdlr`

            :meth:`pygogo.Gogo.update_hdlr`

            :meth:`pygogo.Gogo.zip`

        Returns:
            New instance of :class:`logging.Logger`

        Examples:
            >>> going = Gogo()
            >>> logger = going.get_logger('default')
            >>> logger # doctest: +ELLIPSIS
            <...Logger...>
            >>> logger.info('from default')
            from default
            >>> going.get_logger('new').info('from new')
            from new
        """
        lggr_name = '%s.%s' % (self.name, name)
        logger = logging.getLogger(lggr_name)

        if lggr_name not in self.loggers:
            self.loggers.add(lggr_name)

            if kwargs:
                kwargs['name'] = lggr_name

            fmtrs = [self.formatters['high'], self.formatters['low']]

            for zipped in self.zip(*fmtrs):
                hdlr, level, fmtr, monolog = zipped
                copied_hdlr = copy_hdlr(hdlr)
                self.update_hdlr(copied_hdlr, level, fmtr, monolog, **kwargs)
                logger.addHandler(copied_hdlr)

            logger.setLevel(self.levels['low'])

        return logger

    def get_structured_logger(self, name=None, **kwargs):
        """Retrieve a structured data logger

        Args:
            name (string): The logger name.

            kwargs (dict): Keyword arguments to include in every log message.

        See also:
            :func:`pygogo.copy_hdlr`

            :meth:`pygogo.Gogo.update_hdlr`

            :meth:`pygogo.Gogo.zip`

            :class:`pygogo.utils.StructuredAdapter`

        Returns:
            New instance of :class:`pygogo.utils.StructuredAdapter`

        Examples
            >>> from io import StringIO
            >>> from json import loads

            >>> s = StringIO()
            >>> going = Gogo('structured', low_hdlr=handlers.fileobj_hdlr(s))
            >>> logger = going.get_structured_logger(all='true')
            >>> logger  # doctest: +ELLIPSIS
            <...StructuredAdapter...>
            >>> logger.debug('hello')
            >>> logger.debug('extra', extra={'key': 'value'})
            >>> s.seek(0) or 0
            0
            >>> loads(next(s)) == {'all': 'true', 'message': 'hello'}
            True
            >>> loads(next(s)) == {
            ...     'all': 'true', 'message': 'extra', 'key': 'value'}
            True
        """
        # pylint: disable=dict-items-not-iterating
        values = frozenset(kwargs.items())
        name = name or hashlib.md5(str(values).encode('utf-8')).hexdigest()
        lggr_name = '%s.structured.%s' % (self.name, name)
        logger = logging.getLogger(lggr_name)

        if lggr_name not in self.loggers:
            self.loggers.add(lggr_name)
            formatter = formatters.basic_formatter

            for zipped in self.zip(formatter, formatter):
                hdlr, level, fmtr, monolog = zipped
                copied_hdlr = copy_hdlr(hdlr)
                self.update_hdlr(copied_hdlr, level, fmtr, monolog)
                logger.addHandler(copied_hdlr)

            logger.setLevel(self.levels['low'])

        return utils.StructuredAdapter(logger, kwargs)


def copy_hdlr(hdlr):
    """Safely copy a handler and its associated filters.

    Args:
        hdlr (obj): A :class:`logging.handlers` instance.

    See also:
        :meth:`pygogo.Gogo.get_logger`

        :meth:`pygogo.Gogo.get_structured_logger`

    Returns:
        New instance of :class:`logging.handlers`

    Examples:
        >>> hdlr = logging.StreamHandler(sys.stdout)
        >>> copy_hdlr(hdlr) # doctest: +ELLIPSIS
        <...StreamHandler...>
    """
    copied_hdlr = copy(hdlr)
    copied_hdlr.filters = [copy(f) for f in hdlr.filters]
    return copied_hdlr

logger = Gogo().logger
